#define detect 3        //детектор нуля
#define triac 5          //тиристор
#define accuracy 1000L							  // Точность измерения токов после запятой (Больше точнее)
#define zerooffset 10L							  // Сдвиг срабатывания детектора нуля относительно реального нуля в тиках таймера
#define c_timer 625L-zerooffset	          				  // Максимальное количество срабатываний таймера за полупериод синусоиды
#define c_halfcycle 10							  // Количество полупериодов для расчета среднеквадратичного значения тока
#include <ReadDigKey.h>
ReadDigKey key;
#include <Wire.h>
#include <LiquidCrystal_I2C.h> //https://github.com/marcoschwartz/LiquidCrystal_I2C

LiquidCrystal_I2C lcd(0x27, 16, 2);

#define pSW 7			// энкодер SW кнопка
#define pCLK 8                  // Pin для энкодера CLK
#define pDT 9			// Pin для энкодера DT

#define prn_serial		// Закоментарить если не нужно выводить в последовательный порт результат

#define ACS_coeff 0.048828125	// Коэффициент датчика ACS712 |5А - 0.024414063 | 20А - 0.048828125 | 30A - 0.073242188 |
#define max_I 16		//максимальный ток А
#define step_I 0.1		//шаг установки тока А

volatile unsigned int cntr = 0;			// счетчик количества измерений мгновенного значения тока
volatile unsigned int zero = 0;			// Счетчик полупериодов
volatile unsigned long Isumm = 0;		// переменная для хранения суммы квадратов
volatile unsigned int tstamp_lcd = 0;		// Количество периодов между выводами в дисплей и serial port

unsigned long ust_I = 0;			// установленный ток
unsigned long real_I = 0;			// переменная расчета RMS тока
unsigned long angle_I = (c_timer) * accuracy;

long buffer_ust_I = 0;
bool set_menu_I = false;

//======================================================== Прерывание по нулю
void zero_crosss_int() {
	TCNT1 = 0;
	PORTD &= ~(1 << PORTD5);
	zero++;
	tstamp_lcd++;
}
//======================================================== Обработка таймера по совпадению нуля
ISR(TIMER1_COMPA_vect) {
	if (TCNT1 < c_timer) PORTD |= (1 << PORTD5);
}

//======================================================== Обработка прерывания АЦП для расчета среднеквадратичного тока
ISR(ADC_vect) {
	unsigned long Iism = 0;	//мгновенные значения тока
	byte An_pin = ADCL;
	byte An = ADCH;
	Iism = ((An << 8) + An_pin) - 512;
	Iism *= Iism;                    // возводим значение в квадрат
	Isumm += Iism;                   // складываем квадраты измерений
	cntr++;
}

void setup(void) {
#ifdef prn_serial
	Serial.begin(115200);		 //активируем прием сообщения
#endif
	pinMode(detect, INPUT);          //детектор нуля
	pinMode(triac, OUTPUT);          //тиристор
	PORTD &= ~(1 << PORTD5);
	key.add_key(pSW);
	key.add_enccoder(pCLK, pDT);
	lcd.init();
	lcd.backlight();
	//===========================================================Настройка АЦП
	ADMUX = (0 << REFS1) | (1 << REFS0) | (0 << MUX2) | (0 << MUX1) | (1 << MUX0); //
	ADCSRA = B11101111; //Включение АЦП
	ACSR = (1 << ACD);
	//------ Timer1 ---------- Таймер задержки времени открытия триака после детектирования нуля (0 триак не откроется)
	TCCR1A = 0x00;  //
	TCCR1B = 0x00;    //
	TCCR1B = (1 << CS12) | (0 << CS11) | (0 << CS10); // Тактирование от CLK.
	OCR1A = 0;                   // Верхняя граница счета. Диапазон от 0 до 65535.
	TIMSK1 |= (1 << OCIE1A);     // Разрешить прерывание по совпадению
	attachInterrupt(1, zero_crosss_int, RISING);//вызов прерывания при детектировании нуля
}

void loop(void) {
	if (zero >= c_halfcycle) {	          					//расчет среднеквадратичного
		zero = 0;
		real_I = sqrt(Isumm / cntr) * ACS_coeff * double(accuracy);
		Isumm = cntr = 0;							// обнуляем суммы токов и счетчики
	}
	if (ust_I) {									// Расчет угла открытия триака
		angle_I += real_I - ust_I;
		if (angle_I < (zerooffset)*accuracy) angle_I = (zerooffset)*accuracy;
		if (angle_I > (c_timer)*accuracy) angle_I = (c_timer)*accuracy;
	} else
		angle_I = (c_timer)*accuracy;
	OCR1A = int(angle_I / accuracy);
	delay(1);									// Задержка чтобы не попасть в резонанс программы и прерываний

	key.readkey();
	if (key.shot_press()) {			        				// Короткое нажатие кнопки
		if (set_menu_I) {							// Выйти из меню без сохранения
			buffer_ust_I = ust_I;
			set_menu_I = false;
		} else {
			set_menu_I = true;
		}
	} else if (key.long_press()) {							// Долгое нажатие кнопки
		if (set_menu_I) {							// Выйти и сохранить значение
			ust_I = buffer_ust_I;
			set_menu_I = false;
		}
	}
	// Энкодер
	key.readencoder();
	if (key.encoder) {				 // вращение
		set_menu_I = true;
		buffer_ust_I += step_I * key.encoder * accuracy;
		if (buffer_ust_I > max_I * accuracy) buffer_ust_I = 0;
		if (buffer_ust_I < 0) buffer_ust_I = max_I * accuracy;
		key.encoder = 0;
	}

	while (Serial.available()) {               //Serial port, пока не конец сообщения, читаем данные и формируем строку
		static String buf_string="";
		char ch = Serial.read();
		buf_string += ch;
		if (ch == '\n') {
			if (buf_string.substring(0, 2) == "TW") {
				String T1 = buf_string.substring(buf_string.indexOf("TW", 2) + 3);
				ust_I = T1.toFloat() * float(accuracy);       //Выставленный ток с Serial порта
				buffer_ust_I = ust_I;
			}
			buf_string = "";
			exit;
		}
	}
	//=========================Информация в порт и дисплей
	char buff[5];
	String str0, str1;
	if (tstamp_lcd  >= 30) {
		tstamp_lcd = 0;
		dtostrf(float(real_I) / accuracy, 5, 2, buff);
		str0 = "Curr:" + String(buff);
		dtostrf(float(ust_I) / accuracy, 5, 2, buff);
		str1 = "Inst:" + String(buff);
		lcd.setCursor(0, 0); lcd.print(str0);
		lcd.setCursor(0, 1); lcd.print(str1);
		if (set_menu_I) {
			dtostrf(float(buffer_ust_I) / accuracy, 5, 2, buff);
			lcd.print(" "+String(buff));
		} else
			lcd.print("      ");
#ifdef prn_serial
		Serial.println(str0);				Serial.println(str1);
		Serial.println("-------------");
#endif
	}
}







